/*
MIT License

Copyright (c) 2022 Jason Turner

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

// Important note: the types in this file are only intended for compile-time construction
// but consteval doesn't exist in C++17

#ifndef CONSTEXPR_JSON_HPP_INCLUDED
#define CONSTEXPR_JSON_HPP_INCLUDED

#include <array>
#include <cassert>
#include <cstdint>
#include <stdexcept>
#include <string_view>

namespace constexpr_json {
template<typename First, typename Second> struct pair
{
  First first;
  Second second;
};

template<typename T> struct span
{
  template<std::size_t Size>
  constexpr explicit span(const std::array<T, Size> &input)
    : begin_{ input.data() }, end_{ std::next(input.data(), Size) }
  {}

  constexpr span() : begin_{ nullptr }, end_{ nullptr } {}

  constexpr const T *begin() const { return begin_; }

  constexpr const T *end() const { return end_; }

  constexpr std::size_t size() const { return static_cast<std::size_t>(std::distance(begin_, end_)); }

  const T *begin_;
  const T *end_;
};

template<typename CharType> struct basic_json;
template<typename CharType> using basic_array_t = span<basic_json<CharType>>;
template<typename CharType> using basic_value_pair_t = pair<std::basic_string_view<CharType>, basic_json<CharType>>;
template<typename CharType> using basic_object_t = span<basic_value_pair_t<CharType>>;

using binary_t = span<std::uint8_t>;

template<typename CharType> struct data_variant
{
  struct monostate
  {
  };

  union value_t {
    monostate empty_;
    bool bool_;
    binary_t binary_;
    basic_array_t<CharType> array_;
    basic_object_t<CharType> object_;
    std::int64_t int64_t_;
    std::uint64_t uint64_t_;
    double double_;
    std::basic_string_view<CharType> string_view_;

    constexpr explicit value_t() : empty_{} {}
    constexpr explicit value_t(monostate) : value_t() {}
    constexpr explicit value_t(bool b) : bool_{ b } {}
    constexpr explicit value_t(binary_t b) : binary_{ b } {}
    constexpr explicit value_t(basic_array_t<CharType> a) : array_{ a } {}
    constexpr explicit value_t(basic_object_t<CharType> o) : object_{ o } {}
    constexpr explicit value_t(std::int64_t i) : int64_t_{ i } {}
    constexpr explicit value_t(std::uint64_t i) : uint64_t_{ i } {}
    constexpr explicit value_t(double d) : double_{ d } {}
    constexpr explicit value_t(std::basic_string_view<CharType> s) : string_view_{ s } {}
  };

  enum struct selected_type { empty, boolean, binary, array, object, integer, uinteger, floating_point, string };

  value_t value{ monostate{} };
  selected_type selected{ selected_type::empty };

  // letting these be implicit measurably improves construction performance

  constexpr data_variant() = default;
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(monostate) : data_variant() {}
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(bool b) : value{ b }, selected{ selected_type::boolean } {}
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(basic_array_t<CharType> a) : value{ a }, selected{ selected_type::array } {}
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(basic_object_t<CharType> o) : value{ o }, selected{ selected_type::object } {}
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(std::int64_t i) : value{ i }, selected{ selected_type::integer } {}
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(std::uint64_t i) : value{ i }, selected{ selected_type::uinteger } {}
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(double d) : value{ d }, selected{ selected_type::floating_point } {}
  // cppcheck-suppress noExplicitConstructor
  constexpr data_variant(std::basic_string_view<CharType> s) : value{ s }, selected{ selected_type::string } {}

  [[nodiscard]] constexpr const bool *get_if_boolean() const noexcept
  {
    if (selected == selected_type::boolean) {
      return &value.bool_;
    } else {
      return nullptr;
    }
  }

  [[nodiscard]] constexpr const basic_array_t<CharType> *get_if_array() const noexcept
  {
    if (selected == selected_type::array) {
      return &value.array_;
    } else {
      return nullptr;
    }
  }
  [[nodiscard]] constexpr const basic_object_t<CharType> *get_if_object() const noexcept
  {
    if (selected == selected_type::object) {
      return &value.object_;
    } else {
      return nullptr;
    }
  }
  [[nodiscard]] constexpr const std::int64_t *get_if_integer() const noexcept
  {
    if (selected == selected_type::integer) {
      return &value.int64_t_;
    } else {
      return nullptr;
    }
  }
  [[nodiscard]] constexpr const std::uint64_t *get_if_uinteger() const noexcept
  {
    if (selected == selected_type::uinteger) {
      return &value.uint64_t_;
    } else {
      return nullptr;
    }
  }

  [[nodiscard]] constexpr const double *get_if_floating_point() const noexcept
  {
    if (selected == selected_type::floating_point) {
      return &value.double_;
    } else {
      return nullptr;
    }
  }

  [[nodiscard]] constexpr const std::basic_string_view<CharType> *get_if_string() const noexcept
  {
    if (selected == selected_type::string) {
      return &value.string_view_;
    } else {
      return nullptr;
    }
  }
};

template<typename CharType> struct basic_json
{
  using data_t = data_variant<CharType>;

  struct iterator
  {
    constexpr explicit iterator(const basic_json &value, std::size_t index = 0) : parent_value_(&value), index_{ index }
    {}

    constexpr const basic_json &operator*() const
    {
      if (parent_value_->is_array()) {
        return (*parent_value_)[index_];
      } else if (parent_value_->is_object()) {
        return std::next(parent_value_->object_data().begin(), static_cast<std::ptrdiff_t>(index_))->second;
      } else {
        return *parent_value_;
      }
    }

    constexpr const basic_json *operator->() const { return &(*(*this)); }

    constexpr std::size_t index() const { return index_; }

    constexpr const basic_json &value() const { return *(*this); }


    constexpr std::basic_string_view<CharType> key() const
    {
      if (parent_value_->is_object()) {
        return std::next(parent_value_->object_data().begin(), static_cast<std::ptrdiff_t>(index_))->first;
      } else {
        throw std::runtime_error("json value is not an object, it has no key");
      }
    }

    constexpr bool operator==(const iterator &other) const
    {
      return other.parent_value_ == parent_value_ && other.index_ == index_;
    }
    constexpr bool operator!=(const iterator &other) const { return !(*this == other); }


    constexpr bool operator<(const iterator &other) const
    {
      assert(other.parent_value_ == parent_value_);
      return index_ < other.index_;
    }

    constexpr iterator &operator--()
    {
      --index_;
      return *this;
    }

    constexpr iterator operator--(int)
    {
      iterator result{ *this };
      index_--;
      return result;
    }

    constexpr iterator &operator++()
    {
      ++index_;
      return *this;
    }

    constexpr iterator operator++(int)
    {
      iterator result{ *this };
      index_++;
      return result;
    }

    constexpr iterator &operator+=(const std::ptrdiff_t value)
    {
      index_ = static_cast<std::size_t>(static_cast<std::ptrdiff_t>(index_) + value);
      return *this;
    }


    constexpr iterator &operator+=(const std::size_t value)
    {
      index_ += value;
      return *this;
    }

    const basic_json *parent_value_{ nullptr };
    std::size_t index_{ 0 };
  };

  using const_iterator = iterator;

  constexpr iterator begin() const { return iterator{ *this }; }

  constexpr iterator end() const { return iterator{ *this, size() }; }

  constexpr iterator cbegin() const { return begin(); }

  constexpr iterator cend() const { return end(); }

  constexpr std::size_t size() const noexcept
  {
    if (is_null()) { return 0; }
    if (is_object()) { return data.get_if_object()->size(); }

    if (is_array()) { return data.get_if_array()->size(); }

    return 1;
  }

  constexpr const basic_json &operator[](const std::size_t idx) const
  {
    if (const auto &children = array_data(); idx < children.size()) {
      return *std::next(children.begin(), static_cast<std::ptrdiff_t>(idx));
    } else {
      throw std::runtime_error("index out of range");
    }
  }

  constexpr iterator find(const std::basic_string_view<CharType> key) const
  {
    for (auto itr = begin(); itr != end(); ++itr) {
      if (itr.key() == key) { return itr; }
    }

    return end();
  }

  constexpr const basic_json &operator[](const std::basic_string_view<CharType> key) const
  {
    const auto &children = object_data();

    // find_if is not constexpr in C++17, so we rolled our own,
    // and this helps us work around bugs in older versions of GCC
    // and constexpr
    const auto finder = [&]() {
      auto itr = children.begin();

      for (; itr != children.end(); ++itr) {
        if (itr->first == key) { return itr; }
      }

      return itr;
    };

    const auto obj = finder();

    if (obj != children.end()) {
      return obj->second;
    } else {
      throw std::runtime_error("Key not found");
    }
  }

  constexpr const auto &array_data() const
  {
    if (const auto *result = data.get_if_array(); result != nullptr) {
      return *result;
    } else {
      throw std::runtime_error("value is not an array type");
    }
  }

  constexpr const auto &object_data() const
  {
    if (const auto *result = data.get_if_object(); result != nullptr) {
      return *result;
    } else {
      throw std::runtime_error("value is not an object type");
    }
  }

  constexpr static basic_json object() { return basic_json{ data_t{ basic_object_t<CharType>{} } }; }
  constexpr static basic_json array() { return basic_json{ data_t{ basic_array_t<CharType>{} } }; }

  template<typename Type> constexpr Type get() const
  {
    if constexpr (std::is_same_v<Type, std::uint64_t> || std::is_same_v<Type, std::int64_t>) {
      if (const auto *uint_value = data.get_if_uinteger(); uint_value != nullptr) {
        return Type(*uint_value);
      } else if (const auto *value = data.get_if_integer(); value != nullptr) {
        return Type(*value);
      }
    } else if constexpr (std::is_same_v<Type, double>) {
      if (const auto *value = data.get_if_floating_point(); value != nullptr) { return *value; }
    } else if constexpr (std::is_same_v<Type,
                           std::basic_string_view<CharType>> || std::is_same_v<Type, std::basic_string<CharType>>) {
      if (const auto *value = data.get_if_string(); value != nullptr) { return *value; }
    } else if constexpr (std::is_same_v<Type, bool>) {
      if (const auto *value = data.get_if_boolean(); value != nullptr) { return *value; }
    } else {
      throw std::runtime_error("Unexpected type for get()");
    }

    throw std::runtime_error("Incorrect type for get()");
  }

  constexpr bool is_object() const noexcept { return data.selected == data_t::selected_type::object; }

  constexpr bool is_array() const noexcept { return data.selected == data_t::selected_type::array; }

  constexpr bool is_string() const noexcept { return data.selected == data_t::selected_type::string; }

  constexpr bool is_boolean() const noexcept { return data.selected == data_t::selected_type::boolean; }

  constexpr bool is_structured() const noexcept { return is_object() || is_array(); }

  constexpr bool is_number() const noexcept { return is_number_integer() || is_number_float(); }

  constexpr double as_number_float() const
  {
    if (const double *value = data.get_if_floating_point(); value != nullptr) {
      return *value;
    } else {
      throw std::runtime_error("Not a float type");
    }
  }
  constexpr double as_boolean() const
  {
    if (const bool *value = data.get_if_boolean(); value != nullptr) {
      return *value;
    } else {
      throw std::runtime_error("Not a boolean type");
    }
  }

  constexpr auto as_string() const
  {
    if (const auto *value = data.get_if_string(); value != nullptr) {
      return *value;
    } else {
      throw std::runtime_error("Not a string type");
    }
  }

  constexpr operator double() const { return as_number_float(); }

  constexpr bool is_number_integer() const noexcept { return is_number_signed() || is_number_unsigned(); }

  constexpr bool is_number_signed() const noexcept { return data.selected == data_t::selected_type::integer; }

  constexpr bool is_number_unsigned() const noexcept { return data.selected == data_t::selected_type::uinteger; }

  constexpr bool is_number_float() const noexcept { return data.selected == data_t::selected_type::floating_point; }

  constexpr bool is_null() const noexcept { return data.selected == data_t::selected_type::empty; }

  constexpr bool is_binary() const noexcept { return data.selected == data_t::selected_type::binary; }

  constexpr bool is_primitive() const noexcept
  {
    return is_null() || is_string() || is_boolean() || is_number() || is_binary();
  }

  data_t data;
};

using json = basic_json<char>;
using object_t = basic_object_t<char>;
using value_pair_t = basic_value_pair_t<char>;
using array_t = basic_array_t<char>;
}// namespace constexpr_json

#endif
